use crate::radio::MAX_PACKET_LEN;
use colored::{Color, Colorize};
use ham_cats::{
    identity::Identity,
    packet::Packet,
    whisker::{PastHop, Route, RouteHop},
};

pub fn print_packet(pkt: &Packet<MAX_PACKET_LEN>, rssi: Option<f64>) {
    println!();

    let color = match rssi {
        Some(rssi) => {
            println!("----- RF [{rssi} dBm] -----");

            Color::Green
        }
        None => {
            println!("--- FELINET ---");

            Color::Magenta
        }
    };

    if let Some(ident) = pkt.identification() {
        println!(
            "{} \t{}-{} [ICON: {}]",
            "IDENT:".color(color),
            ident.callsign,
            ident.ssid,
            ident.icon
        );
    }
    if let Some(route) = pkt.route() {
        let r_str = route
            .iter()
            .map(|rn| match rn {
                RouteHop::Internet => "[NET]".to_string(),
                RouteHop::Past(past_hop) => {
                    let ident = past_hop.identity();
                    let rc = ident.callsign();
                    let rs = ident.ssid();
                    let rssi = match past_hop.rssi() {
                        Some(x) => format!(" [{x:.1} dbm]"),
                        None => "".to_string(),
                    };
                    format!("{rc}-{rs}{rssi}")
                }
                RouteHop::Future(ident) => {
                    let rc = ident.callsign();
                    let rs = ident.ssid();
                    format!("{rc}-{rs}*")
                }
            })
            .collect::<Vec<String>>()
            .join(" -> ");
        println!(
            "{} \t(MAX {}) {r_str}",
            "ROUTE:".color(color),
            route.max_hops
        );
    }

    let mut buf = [0; MAX_PACKET_LEN];
    if let Ok(comment) = pkt.comment(&mut buf) {
        println!("{} \t'{comment}'", "CMNT:".color(color));
    }
    if let Some(gps) = pkt.gps() {
        println!(
            "{} \t({:07.4}, {:07.4}) +/- {:03} m, v = {:.2} m/s [N {:.2} deg]",
            "GPS:".color(color),
            gps.latitude(),
            gps.longitude(),
            gps.max_error,
            gps.speed,
            gps.heading()
        );
    }

    if let Some(info) = pkt.node_info() {
        match (info.hardware_id(), info.software_id()) {
            (Some(hw), Some(sw)) => {
                println!(
                    "{} \t{hw:#06x} {} {sw:#04x}",
                    "HW:".color(color),
                    "SW:".color(color)
                );
            }
            (Some(hw), None) => {
                println!("{} \t{hw:#06x}", "SW:".color(color));
            }
            (None, Some(sw)) => {
                println!("{} \t{sw:#04x}", "SW:".color(color));
            }
            (None, None) => {}
        }

        if let Some(uptime) = info.uptime() {
            println!("{} \t{} s", "UTIME:".color(color), uptime);
        }

        if let Some(a_height) = info.antenna_height() {
            println!("{} \t{} m", "VERT:".color(color), a_height);
        }

        if let Some(a_gain) = info.antenna_gain() {
            println!("{} \t{} dBi", "GAIN:".color(color), a_gain);
        }

        if let Some(power) = info.tx_power() {
            println!("{} \t{} dBm", "TXP:".color(color), power);
        }

        if let Some(volts) = info.voltage() {
            println!("{} \t{} V", "VOLTS:".color(color), volts);
        }

        if let Some(temp) = info.xcvr_temperature() {
            println!("{} \t{} C", "TEMP:".color(color), temp);
        }
    }
}

pub fn attempt_decode<'a>(
    packet: &[u8],
    buf: &'a mut [u8; MAX_PACKET_LEN],
    rssi: f64,
    debug: bool,
) -> Option<Packet<'a, MAX_PACKET_LEN>> {
    if debug {
        eprintln!("Attempt decode with len: {}", packet.len());
    }
    let packet: Packet<MAX_PACKET_LEN> = match Packet::fully_decode(packet, buf) {
        Ok(x) => x,
        Err(e) => {
            if debug {
                eprintln!("Could not decode packet: {e}");
            }

            return None;
        }
    };

    print_packet(&packet, Some(rssi));

    Some(packet)
}

// i-gating is kind of weird.
// if a packet has a pre-determined route (is_future=true) and our node
// doesn't match it, we don't digipeat it.
// But FELINET should be an "all-seeing eye". Any packet that's heard
// should end up on the network.
// The solution is that, if the packet has non-matching is_future=true nodes,
// We rip them off, and stick our node + an internet route on
#[must_use]
pub fn append_internet_to_packet_route(
    callsign: &str,
    ssid: u8,
    rssi: f64,
    packet: &mut Packet<MAX_PACKET_LEN>,
) -> Option<()> {
    let mut route = packet.route().unwrap_or(Route::new(0));

    let first_future = route.iter().find_map(|rn| match rn {
        RouteHop::Future(ident) => Some((ident.callsign(), ident.ssid())),
        _ => None,
    });

    if let Some((rc, rs)) = first_future {
        if rc == callsign && rs == ssid {
            // we can just do the normal thing
            // since we're replacing the existing future node, this should never fail, unless we run out of byte space
            route
                .append_hop(PastHop::new(Identity::new(callsign, ssid), Some(rssi)))
                .ok()?;
        } else {
            // rip off all the futures and add ourselves
            let mut new_route = Route::new(route.max_hops);
            for r in route.iter() {
                match r {
                    RouteHop::Internet => new_route.push_internet()?,
                    RouteHop::Past(past) => {
                        new_route.push_past(past)?;
                    }
                    RouteHop::Future(_) => break,
                }
            }

            route = new_route;
            route.push_past(PastHop::new(Identity::new(callsign, ssid), Some(rssi)))?;
        }
    } else {
        route.push_past(PastHop::new(Identity::new(callsign, ssid), Some(rssi)))?;
    }

    route.push_internet()?;
    packet.clear_route();
    packet.add_route(route).ok()?;

    Some(())
}

pub fn validate_freq(f: u32) -> bool {
    (420_000_000..=450_000_000).contains(&f)
}
